/*
 * oscp-slave-ffm.c
 * Copyright (C) 2015 KotCzarny <tjosko@yahoo.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "oscp-slave.h"
#include <ctype.h>

#include "libavcodec/avcodec.h"
#include "libavutil/avutil.h"
#include "libavutil/samplefmt.h"
#include "libavutil/opt.h"
#include "libavformat/avformat.h"
#include "libswscale/swscale.h"
#ifdef FFM
#include "libswresample/swresample.h" //ffmpeg
#else
#include "libavresample/avresample.h" //libav
#endif

int get_mp3_duration(char *fn);

//#define AVCODEC_MAX_AUDIO_FRAME_SIZE 192000/16
#define AVCODEC_MAX_AUDIO_FRAME_SIZE 16384

//SwrContext *swr;

AVFormatContext* container; AVDictionary *metadata; AVCodecContext *ctx; AVCodec *codec;
AVPacket packet; AVFrame *ffmframe;
enum AVSampleFormat sfmt;

int ffmbuffer_size, ffmstream_id=-1; int ffmstate=0; float ffmcur=0, ffmdur=0; int ffmi=0;
char ffmfn[1024];

/////###############
void ffm_show_supp() { av_register_all();
 printf("  [slave-libav]: based on libav-1.1");
 printf(", formats: mp3, wav, ogg, flac, ape, mpg, mpeg, mpc, aac, mp2, wma, wmv, wv, asf, rm\n");

// printf("\n codecs:"); AVCodec * codec = av_codec_next(NULL); while(codec != NULL)
// { printf(" %s", codec->name); codec = av_codec_next(codec); } printf("\n");
// printf("\n iformats:"); AVInputFormat * iformat = av_iformat_next(NULL); while(iformat != NULL)
// { printf(" %s", iformat->name); iformat = av_iformat_next(iformat); } printf("\n");
// printf("\n oformats:"); AVOutputFormat * oformat = av_oformat_next(NULL); while(oformat != NULL)
// { printf(" %s", oformat->name); oformat = av_oformat_next(oformat); } printf("\n");
}


//#define audio_outbuf_size 44100*4
//static uint16_t audio_outbuf[audio_outbuf_size];


/////###########
//void ffm_flush() { lf("ffm_flush"); adev_flush(); }
/////##########
void ffm_init() { lf("ffm_init"); ffmstate=0; dlog(9, "ffm_init"); if(ffmstate>0) return;
 av_register_all(); av_log_set_level(AV_LOG_ERROR); container=avformat_alloc_context();

 if(adev_init()<=0) { ffmstate=0; return; }

 ffmstate=1;
}
/////##########
void ffm_stop() { lf("ffm_stop");
 dlog(9, "ffm_stop");			 if(packet.data) av_free_packet(&packet);
 if(0)cs("ffm_stop: avcodec_close\n");	 if(ctx) avcodec_close(ctx);
 if(0)cs("ffm_stop: avformat_close_input\n"); if(container) avformat_close_input(&container); 
 if(0)cs("ffm_stop: done\n");		 ffmstate=1;
}
/////#########
void ffm_end() { lf("ffm_end");
 dlog(9, "ffm_end"); if(ffmstate>1) ffm_stop(); adev_end();
 ffmstate=0; 
}
/////##################
void ffm_die(char *msg) { lf("ffm_die"); dlog(9, "ffm_die"); ffm_stop(); cs(msg); }

/////#################
void ffm_update_info() { static char lmsg[4096], keyl[16], *p=keyl;
 int title=0, artist=0;
 AVDictionaryEntry *tag = NULL; while((tag=av_dict_get(metadata, "", tag, AV_DICT_IGNORE_SUFFIX)))
 {
  strncpy(keyl, tag->key, 15); p=keyl; for ( ; *p; ++p) *p = tolower(*p);

  if(0==strcmp(keyl,"artist")) { artist++; snprintf(lmsg,1022,"Artist: %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"title")) { title++; snprintf(lmsg,1022,"Title: %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"album")) { snprintf(lmsg,1022,"Album: %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"comment")) { snprintf(lmsg,1022,"Comment: %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"genre")) { snprintf(lmsg,1022,"Genre  : %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"date")) { snprintf(lmsg,1022,"Year: %s\n",tag->value); cs(lmsg); }

  else if(0==strcmp(keyl,"StreamTitle")) { snprintf(lmsg,1022,"Title: %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"icy-genre")) { snprintf(lmsg,1022,"Genre  : %s\n",tag->value); cs(lmsg); }
  else if(0==strcmp(keyl,"icy-name")) { snprintf(lmsg,1022,"Comment: %s\n",tag->value); cs(lmsg); }
snprintf(lmsg, 1022, "ffm_update_info: %s=%s", tag->key, tag->value); dlog(9,lmsg);
/* FIXME: handle:
[10 9 ffm_update_info: icy-notice1=<BR>This stream requires <a href="http://www.winamp.com">Winamp</a><BR>]
[10 9 ffm_update_info: icy-notice2=SHOUTcast DNAS/posix(linux x64) v2.4.2.167<BR>]
[10 9 ffm_update_info: icy-name=MB RECASTER]
[10 9 ffm_update_info: icy-genre=MB RECASTER]
[10 9 ffm_update_info: icy-url=http://www.mbradio.it]
[10 9 ffm_update_info: icy-pub=0]
[10 9 ffm_update_info: icy-br=128]

[36 9 ffm_update_info: icy-notice1=<BR>This stream requires <a href="http://www.winamp.com/">Winamp</a><BR>]
[36 9 ffm_update_info: icy-notice2=SHOUTcast Distributed Network Audio Server/Linux v1.9.8<BR>]
[36 9 ffm_update_info: icy-name=Multiradio]
[36 9 ffm_update_info: icy-genre=Adult Contemporary]
[36 9 ffm_update_info: icy-url=http://www.multiradio.it]
[36 9 ffm_update_info: icy-pub=1]
[36 9 ffm_update_info: icy-br=64]
[36 9 ffm_update_info: StreamTitle=J.Ax - uno di questi giorni]
[36 9 ffm_update_info: StreamUrl=]
[36 9 ffm_play]
*/
 } container->event_flags&= !AVFMT_EVENT_FLAG_METADATA_UPDATED;
 if(artist==0 && title==0) { snprintf(lmsg,1023,"Title: %s\n",get_basename(ffmfn)); cs(lmsg); }
}

void ffm_play();
int oscp_pls2link(char *url, char *buf); // buf must be allocated and 4096
/////###############################
void ffm_load(char *input_filename0) { lf("ffm_load");
 static char lmsg[4096]; lmsg[4095]=0;
 snprintf(lmsg, 4095, "Playing %s\n", input_filename0); cs(lmsg);

 if(input_filename0[0]==0) return;
 if(input_filename0[0]==*"\"" && input_filename0[1]==*"\"" && input_filename0[2]==0)
 { cs("EOF code: stop requested\n"); if(ffmstate>2) ffm_stop(); return; }

 if(ffmstate==0) ffm_init();
 ffm_stop();
 ffmstate=2; ffmcur=0; ffmdur=0;
 int t1,t2;
 char *input_filename=input_filename0;


 // check if input_filename0 is actually a playlist
 if(strncmp("http://", input_filename0, 7)==0) {
  static char stream_url[4096]; stream_url[0]=0;
  int status=oscp_pls2link(input_filename, stream_url);
  if(status<0 && status>=-4) { ffm_die("Error: ffm_load: connect() failed\n"); return; }
  if(stream_url[0]!=0) {
   dlog(9,"its a playlist, switching to first entry");
   input_filename=stream_url;
  } else if(status>=400) { ffm_die("Error: ffm_load: http status >=400\n"); return; }
  else dlog(9,"not a playlist, passing to decoder");
 }

 if(strncmp("mms://", input_filename,6)==0) { // fix mms->mms[th]
  memmove(input_filename+1, input_filename, 4094); input_filename[4095]=0;
  memcpy(input_filename, "mmst", 4);
  if(avformat_open_input(&container,input_filename,NULL,NULL)<0){ input_filename[3]=*"h";
   if(avformat_open_input(&container,input_filename,NULL,NULL)<0){
    memcpy(input_filename, "rtsp", 4);
    if(avformat_open_input(&container,input_filename,NULL,NULL)<0){
     ffm_die("EOF code: Could not open mms://\n"); return; } } }
  dlog(9,input_filename);
 }

// cs("Playing.\n");
// snprintf(lmsg, 4095, "Name   : %s\n", input_filename); cs(lmsg);
// snprintf(lmsg, 4095, "Comment: %s\n", input_filename0); cs(lmsg);


 if(avformat_open_input(&container,input_filename,NULL,NULL)<0){ ffm_die("EOF code: Could not open file\n"); return; }
 if(avformat_find_stream_info(container, NULL)<0){ ffm_die("EOF code: Could not find file info\n"); return; }

 int i; for(i=0;i<container->nb_streams;i++){
  if(container->streams[i]->codec->codec_type==AVMEDIA_TYPE_AUDIO){ ffmstream_id=i; break; } }
 if(ffmstream_id==-1){ ffm_die("EOF code: Could not find Audio Stream\n"); return; }

 metadata=container->metadata;


 ctx=container->streams[ffmstream_id]->codec;
 codec=avcodec_find_decoder(ctx->codec_id);
 if(codec==NULL){ ffm_die("EOF code: cannot find codec\n"); return; }
 if(avcodec_open2(ctx,codec,NULL)<0){ ffm_die("EOF code: Codec cannot be found\n"); return; }

//[AV_SAMPLE_FMT_U8] = { .name = "u8", .bits = 8, .planar = 0, .altform = AV_SAMPLE_FMT_U8P },
//[AV_SAMPLE_FMT_S16] = { .name = "s16", .bits = 16, .planar = 0, .altform = AV_SAMPLE_FMT_S16P },
//[AV_SAMPLE_FMT_S32] = { .name = "s32", .bits = 32, .planar = 0, .altform = AV_SAMPLE_FMT_S32P },
//[AV_SAMPLE_FMT_FLT] = { .name = "flt", .bits = 32, .planar = 0, .altform = AV_SAMPLE_FMT_FLTP },
//[AV_SAMPLE_FMT_DBL] = { .name = "dbl", .bits = 64, .planar = 0, .altform = AV_SAMPLE_FMT_DBLP },
//[AV_SAMPLE_FMT_U8P] = { .name = "u8p", .bits = 8, .planar = 1, .altform = AV_SAMPLE_FMT_U8 },
//[AV_SAMPLE_FMT_S16P] = { .name = "s16p", .bits = 16, .planar = 1, .altform = AV_SAMPLE_FMT_S16 },
//[AV_SAMPLE_FMT_S32P] = { .name = "s32p", .bits = 32, .planar = 1, .altform = AV_SAMPLE_FMT_S32 },
//[AV_SAMPLE_FMT_FLTP] = { .name = "fltp", .bits = 32, .planar = 1, .altform = AV_SAMPLE_FMT_FLT },
//[AV_SAMPLE_FMT_DBLP] = { .name = "dblp", .bits = 64, .planar = 1, .altform = AV_SAMPLE_FMT_DBL },
 sfmt=ctx->sample_fmt;
      if(sfmt==AV_SAMPLE_FMT_U8 || sfmt==AV_SAMPLE_FMT_U8P){ sformat.bits=8; }
 else if(sfmt==AV_SAMPLE_FMT_S16 || sfmt==AV_SAMPLE_FMT_S16P){ sformat.bits=16; }
 else if(sfmt==AV_SAMPLE_FMT_S32 || sfmt==AV_SAMPLE_FMT_S32P){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_FLT || sfmt==AV_SAMPLE_FMT_FLTP){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_DBL || sfmt==AV_SAMPLE_FMT_DBLP){ sformat.bits=64; }
 else { ffm_die("EOF code: Unsupported format\n"); return; }

 sformat.channels=ctx->channels; sformat.rate=ctx->sample_rate;
#ifdef LAO
 sformat.byte_format=AO_FMT_NATIVE; sformat.matrix=0;
#endif

 char *extp=rindex(input_filename, *".");
 if(extp && strcasecmp(extp, ".mp3")==0) {
  ffmdur=get_mp3_duration(input_filename);
 } else {
  t1 = container->streams[ffmstream_id]->duration;
  t2 = av_rescale_q(t1, container->streams[ffmstream_id]->time_base, AV_TIME_BASE_Q);
  ffmdur=t2/1000000.0;
 }
 sprintf(lmsg, "ANS_length=%.1f\n", ffmdur); cs(lmsg);
// sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels,
//  sformat.rate*sformat.channels*sformat.bits/1000); cs(lmsg);
 sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels,
  (ctx->bit_rate)/1000); cs(lmsg);

//   const char* artist = av_dict_get(metadata, "artist", NULL, 0)->value;
//   const char* title = av_dict_get(metadata, "title", NULL, 0)->value;

 strncpy(ffmfn, input_filename, 1023);
 ffm_update_info();

// char ctag[4]; memcpy(ctag, ctx->codec_tag, 4);
 snprintf(lmsg, 80, "Info: [%s,%08x,%s,%d,%d,%d]\n", codec->name,
  ctx->codec_tag, av_get_sample_fmt_name(ctx->sample_fmt), ctx->frame_size,
  ctx->sample_rate, ctx->channels
// ,ctx->time_base.den/ctx->time_base.num == freq

//(int)ctag+3,(int)ctag+2,(int)ctag+1,(int)ctag+0
); cs(lmsg);

 av_init_packet(&packet); ffmframe=avcodec_alloc_frame();
 ffmbuffer_size=AVCODEC_MAX_AUDIO_FRAME_SIZE + FF_INPUT_BUFFER_PADDING_SIZE;
// sprintf(lmsg, "ffmbuffer_size: %d\n", ffmbuffer_size); cs(lmsg);

 adev_setp(sformat.bits, &sformat.channels, &sformat.rate);

/*
 int obits=sformat.bits, ochan=sformat.channels, orate=sformat.rate;
 sprintf(lmsg, "obits: %d, ochan: %d, orate: %d\n", obits, ochan, orate); cs(lmsg);
 // Set up SWR context once you've got codec information
 if(!swr) swr = swr_alloc();
 av_opt_set_int(swr, "in_channel_layout", ctx->channel_layout, 0);
 av_opt_set_int(swr, "out_channel_layout", AV_CH_LAYOUT_STEREO,  0);
 av_opt_set_int(swr, "in_sample_rate", ctx->sample_rate, 0);
 av_opt_set_int(swr, "out_sample_rate", orate, 0);
 av_opt_set_sample_fmt(swr, "in_sample_fmt", ctx->sample_fmt, 0);
 av_opt_set_sample_fmt(swr, "out_sample_fmt", AV_SAMPLE_FMT_S16,  0);
 swr_init(swr);
*/

 ffmstate=3; ffmi=0;
 ffm_play();
}
void oscp_ffmpeg_check_input();
void oscp_ffmpeg_parse(char *a);
/////################# check messages
void ffm_check_input() { lf("ffm_check_input"); static char lmsg[4096]; static int rc;
 static fd_set ffmfsa;  static struct timeval ffmtv,*tvp;
 if(ffmstate==4) { tvp=NULL; }
 else { ffmtv.tv_sec=0; ffmtv.tv_usec=1; tvp=&ffmtv; }

 FD_ZERO(&ffmfsa); if(parent_said>=0) FD_SET(parent_said, &ffmfsa);

 rc=select(FD_SETSIZE, &ffmfsa, NULL, NULL, tvp); if(rc==0) return;
 else if(rc<0) { sprintf(lmsg, "Info: ffmci rc<0, %d, %s\n", errno, strerror(errno)); cs(lmsg); }
 else if(rc>0) // check if input is available
 { fdreadline(parent_said, lmsg, 1020); oscp_ffmpeg_parse(lmsg); }
}
/////##################
void ffm_pause_toggle() { lf("ffm_pause_toggle");
      if(ffmstate==3) { ffmstate=4; cs("===== PAUSE =====\n");
  do { oscp_ffmpeg_check_input(); } while(ffmstate==4);
 }
 else if(ffmstate==4) { ffmstate=3; cs("===== RESUME =====\n"); }
 else cs("Info: pause called but ffmstate != (3 || 4)"); }


const char *av_frame_side_data_name(enum AVFrameSideDataType type)
{
    switch(type) {
    case AV_FRAME_DATA_PANSCAN:         return "AVPanScan";
    case AV_FRAME_DATA_A53_CC:          return "ATSC A53 Part 4 Closed Captions";
    case AV_FRAME_DATA_STEREO3D:        return "Stereoscopic 3d metadata";
    case AV_FRAME_DATA_MATRIXENCODING:  return "AVMatrixEncoding";
    case AV_FRAME_DATA_DOWNMIX_INFO:    return "Metadata relevant to a downmix procedure";
    case AV_FRAME_DATA_REPLAYGAIN:      return "AVReplayGain";
    case AV_FRAME_DATA_DISPLAYMATRIX:   return "3x3 displaymatrix";
    case AV_FRAME_DATA_AFD:		return "AV_FRAME_DATA_AFD";
    }
    return NULL;
}
/////#############
void ffm_checksd() { static int i; static char lmsg[4096];
 if(ffmframe->nb_side_data) {
  for(i=0; i<ffmframe->nb_side_data; i++) {
   AVFrameSideData *sd=ffmframe->side_data[i];
   const char *name=av_frame_side_data_name(sd->type);
   snprintf(lmsg, 4095, "[sd: %s=%d]", name?name:"unknown", sd->size); dlog(9,lmsg);
   }
 }
}
/////#############
void ffm_checkmd() {
 if((container->event_flags & AVFMT_EVENT_FLAG_METADATA_UPDATED)!=0) ffm_update_info();
}
/////##########
void ffm_play() { lf("ffm_play"); dlog(9, "ffm_play");
 if(ffmstate!=3) { cs("EOF code: can't play, not inited\n"); return; }

 static char lmsg[128];
 uint8_t buffer[ffmbuffer_size];  packet.size=ffmbuffer_size;  packet.data=buffer;
 int frameFinished=0, plane_size; int len; //int data_size;
 static int64_t t1, t2, ot=-1;//, ot2;
 int olen=0;

 while(ffmstate==3 && container && av_read_frame(container,&packet)>=0)
 {

  if(packet.stream_index==ffmstream_id){
   avcodec_decode_audio4(ctx,ffmframe,&frameFinished,&packet); // len=
   len=av_samples_get_buffer_size(&plane_size, ctx->channels,
    ffmframe->nb_samples, ctx->sample_fmt, 1); // data_size=
//sprintf(lmsg,"Info: plane_size: %d, len=%d, data_size=%d\n", plane_size, len, data_size); cs(lmsg);
// plane_size: 2304, len=523, data_size=4608

   if(frameFinished){ ffm_checksd();
    ffm_checkmd();

   t1=ffmframe->pkt_dts; if(t1==AV_NOPTS_VALUE) t1=ffmframe->pts;
//    t1 = av_frame_get_best_effort_timestamp(ffmframe);
//t1=0;
    t2 = av_rescale_q(t1, container->streams[ffmstream_id]->time_base, AV_TIME_BASE_Q);
//sprintf(lmsg, "Info: t1: %d, t2: %d\n", (int)t1, (int)t2); cs(lmsg);
    if(t1!= AV_NOPTS_VALUE && t2/1000000 != ot) { /*ot2=1;*/ ot=t2/1000000; ffmcur=t2/1000000.0;
     sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0\n", ffmcur); cs(lmsg);

/*
//     t1 = container->streams[ffmstream_id]->duration;
     t1 = container->duration;
     t2 = av_rescale_q(t1, container->streams[ffmstream_id]->time_base, AV_TIME_BASE_Q);
     ffmdur=t2/1000000.0; sprintf(lmsg, "ANS_length=%.1f\n", ffmdur); cs(lmsg);
*/
    } // else ot2=0;
if(olen!=len) { olen=len;
sprintf(lmsg, "len: %d, plane_size: %d, bits: %d, chan: %d, rate: %d",
 len, plane_size, sformat.bits, sformat.channels, sformat.rate); dlog(9, lmsg); }

//  int swr_convert(struct SwrContext *s, uint8_t **out, int out_count,
// len=swr_convert(swr, audio_outbuf, len, ffmframe->extended_data[0], len);
// adev_out(audio_outbuf, len);

//   adev_out(ffmframe->extended_data, len*8/sformat.bits/ctx->channels, ctx->channels, sfmt);
   adev_out(ffmframe->extended_data, ffmframe->nb_samples, ctx->channels, sfmt);

#if 0
if(0&&ot2==1) { // static char lmsg[1024]; sprintf(lmsg,
 "Info: fmt:%s, ch:%d, bi:%d, ra:%d, len:%d, plane_size: %d",
 lfmt, sformat.channels, sformat.bits, sformat.rate, len, plane_size); dlog(9,lmsg); }
#endif

   } else { cs("frame failed\n"); }
  }
  av_free_packet(&packet);

  ffmi++; ffm_check_input();
 }

 adev_close();
 ffm_die("EOF code: finished decoding\n");
 lf("lfm_play:finished");
}

/////##########################
void oscp_ffmpeg_load(char *fn) { lf("oscp_ffmpeg_load");
 ffm_load(fn); //ffm_play();
 lf("oscp_ffmpeg_load:finished");
}
/////#######################
void oscp_ffmpeg_seek(int a) { lf("oscp_ffmpeg_seek");
 int64_t ts=(int64_t)(a+(int)ffmcur)*1000000;
 ts=av_rescale_q(ts, AV_TIME_BASE_Q, container->streams[ffmstream_id]->time_base);
 if(a<0) avformat_seek_file(container, ffmstream_id, ts,ts,ts, AVSEEK_FLAG_BACKWARD );
 else avformat_seek_file(container, ffmstream_id, ts,ts,ts, 0);

// int64_t seekTime = av_rescale_q(seekTimeDS, ds_time_base, container->streams[ffmstream_id]->time_base);
// int64_t seekStreamDuration = container->streams[ffmstream_id]->duration;
// int flags = AVSEEK_FLAG_BACKWARD; if(seekTime>0 && seekTime<seekStreamDuration)
//  flags |= AVSEEK_FLAG_ANY; // H.264 I frames don't always register as "key frames" in FFmpeg
// int       ret = av_seek_frame(container, ffmstream_id, seekTime, flags);
// if(ret<0) ret = av_seek_frame(container, ffmstream_id, seekTime, AVSEEK_FLAG_ANY);
}
/////##########################
void oscp_ffmpeg_parse(char *a) { lf("oscp_ffmpeg_parse"); char lmsg[4096];
// sprintf(lmsg,"ofparse start"); dlog(9,lmsg);
 int l=strlen(a); if(l<=0) { sprintf(lmsg,"ofparse: l==%d ?", l); dlog(9,lmsg); return; }
 if(strcmp(a,"q")==0 || strcmp(a,"quit")==0) { ffm_end(); exit(0);
 } else if(strncmp(a,"loadfile ",9)==0) { oscp_ffmpeg_load(a+9);
 } else if(strncmp(a,"load ",5)==0) { oscp_ffmpeg_load(a+5);
 } else if(strncmp(a,"pause",5)==0) { ffm_pause_toggle();
 } else if(strncmp(a,"get_property length",19)==0) {
  sprintf(lmsg, "ANS_length=%.1f\n", ffmdur); cs(lmsg);
 } else if(strncmp(a,"get_property time_pos",19)==0 || strncmp(a,"get_time_pos",12)==0) {
//  sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0 0 0\n", ffmcur); cs(lmsg);
 } else if(strncmp(a,"seek ",5)==0) { int o1=0;
  if(a[5]==*"-") { sscanf(a+6, "%d", &o1); o1*=-1; }
  else if(a[5]==*"+") { sscanf(a+6, "%d", &o1); }
  else { sscanf(a+5, "%d", &o1); }
  oscp_ffmpeg_seek(o1);
 } else if(strncmp(a,"vis ",4)==0) { sscanf(a+4, "%d", &vis);
 } else if(strncmp(a,"lfdump",5)==0) { lfdump();
 } else { sprintf(lmsg, "ofparse unhandled: %s", a); dlog(9, lmsg); }
 lf("oscp_ffmpeg_parse:finished");
}
/////######################### check stdin
void oscp_ffmpeg_check_input() { lf("oscp_ffmpeg_check_input");
 char lmsg[4096]; int i, rc; fd_set fsa;
 struct timeval *tvp=NULL; //, tv; tv.tv_sec=0; tv.tv_usec=10000*99; // i=(int)time(NULL);
 usleep(10000); FD_ZERO(&fsa); if(parent_said>=0) FD_SET(parent_said, &fsa);
 rc=select(FD_SETSIZE, &fsa, NULL, NULL, tvp); // block till input

 if(rc<0) { if(0) { perror("select"); exit(EXIT_FAILURE); } else usleep(10000*99); }
 else if(rc==0) { }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==parent_said) { fdreadline(i, lmsg, 1020); oscp_ffmpeg_parse(lmsg); }
 }
 lf("oscp_ffmpeg_check_input:finished");
}
////##################
int oscp_ffmpeg_main() { lf("oscp_ffmpeg_main");
 cs("MPlayer oscp-slave libavcodec "LIBAVCODEC_IDENT"\n");
 lf("oscp_ffmpeg_main:ffm_init");
 ffm_init();
 lf("oscp_ffmpeg_main:while");
 while(1) { oscp_ffmpeg_check_input(); }
 lf("oscp_ffmpeg_main:ffm_end");
 ffm_end();
 lf("oscp_ffmpeg_main:finished");
 cs("EOF code: exit\n");
 return 0;
}
