/*
 * oscp-slave-sid2.c
 * Copyright (C) 2015 KotCzarny <tjosko@yahoo.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "oscp-slave.h"

#include <sidplay/sidplay2.h>
//#include "/1/_src/oscp-cygwin/libs/sidplay-libs-2.1.1/libsidplay/include/sidplay/sidplay2.h"
#include <sidplay/utils/SidTuneMod.h>

void *sid2c; void *sid2t; struct sid2_config_t sid2conf; void *sid2_st; SidTuneInfo sid2_sti;


#define sid2_dbuf_size 1024*2
char sid2_dbuf[sid2_dbuf_size];

int sid2state=0; float sid2cur=0, sid2dur=0; int sid2i=0;
int sid2song=0;

/////################
void sid2_show_supp() {
 printf("  [slave-sid2]: based on sidplay-libs-2.1.1");
 printf(", formats: sid, psid, info\n");
}

/////###########
//void sid2_flush() { lf("sid2_flush"); adev_flush(); }
/////###########
void sid2_init() { lf("sid2_init"); sid2state=0; dlog(9, "sid2_init"); if(sid2state>0) return;

 if(sid2c) dlog(9, "sid2_init: sid2c not null?");

 sid2c=sidplay2_create();
// FILE *fd=fopen("/tmp/sid2.debug", "a"); sidplay2_debug(sid2c, 10, fd);

 if(adev_init()<=0) { sid2state=0; return; }

 sid2state=1;
}
/////##########
void sid2_stop() { lf("sid2_stop");
 dlog(9, "sid2_stop");
 if(sid2state<=1) return;

 if(sid2state>=3) { 
  sidplay2_st_destroy(sid2_st);
 }

 sid2state=1;
}
/////#########
void sid2_end() { lf("sid2_end");
 dlog(9, "sid2_end"); if(sid2state>1) sid2_stop(); adev_end();

 if(sid2c) sidplay2_destroy(sid2c);

 sid2state=0; 
}
/////##################
void sid2_die(char *msg) { lf2("sid2_die",msg); dlog(9, "sid2_die"); sid2_stop(); cs(msg); }

char sid2fn[PATH_MAX]; int sid2fs=0;

float sid2durmax=240;




////#################
int md5toint(char *a) { unsigned int v1, v2, v3, v4; sscanf(a+0, "%8x", &v1);
 sscanf(a+8, "%8x", &v2); sscanf(a+16, "%8x", &v3); sscanf(a+24, "%8x", &v4);
 int hash = v1 ^ v2 ^ v3 ^ v4;
// char lmsg[1024]; snprintf(lmsg, 1023, "md5toint: %s, %8x %8x %8x %8x, hash: %8x", a, v1, v2, v3, v4, hash); dlog(9,lmsg);
 return hash;
}

// #include "oscp-slave-sid2-prepared.c"
int sid2_getdur(int a)
#ifndef SID2GETDUR
{ return 0; }
#endif
;


/////##########################
void sid2_update_info(char *fn) { lf("sid2_update_info");
 sidplay2_st_getinfo(sid2_st, &sid2_sti);
 if(!sid2_sti.formatString) { dlog(9,"empty info ?"); return; }
 static char lmsg[4096]; int fs=get_filesize(fn);
 snprintf(lmsg,1022,"Comment: %d, %s (%s)\n",
  fs, sid2_sti.infoString[2], sid2_sti.formatString); cs(lmsg);

 snprintf(lmsg,1022,"Title: %s (%s)\n", sid2_sti.infoString[0], get_basename(fn)); cs(lmsg);
 snprintf(lmsg,1022,"Artist: %s\n", sid2_sti.infoString[1]); cs(lmsg);

 sid2dur=0;
#if 1
 char md5[36]; sidplay2_stm_createMD5 (sid2_st, md5);
// snprintf(lmsg,1023,"md5: %s, hash: %08x", md5, md5toint(md5)); dlog(9,lmsg);
 md5[0]=97+0;
 int hash=md5toint(md5);
 sid2dur=sid2_getdur(hash);
snprintf(lmsg,1023,"md5: %s, hash: %08x, dur: %d", md5, md5toint(md5), (int)sid2dur); dlog(9,lmsg);
#endif
 if(sid2dur==0) sid2dur=sid2durmax;

 sprintf(lmsg, "ANS_length=%.1f\n", sid2dur); cs(lmsg);
}


void sid2_play();
/////################################
void sid2_load(char *input_filename0) { lfd("sid2_load:state",sid2state);
 static char lmsg[4096]; lmsg[4095]=0;
 snprintf(lmsg, 4095, "Playing %s\n", input_filename0); cs(lmsg);

 if(input_filename0[0]==0) return;
 if(input_filename0[0]==*"\"" && input_filename0[1]==*"\"" && input_filename0[2]==0)
 { cs("EOF code: stop requested\n"); if(sid2state>2) sid2_stop(); return; }

 if(sid2state==0) sid2_init();
 else if(sid2state>2) sid2_stop();

 sid2state=2; sid2cur=0; sid2dur=0;
 char *input_filename=input_filename0;
 int ret=0;


 sid2fs=get_filesize(input_filename);
 if(sid2fs<=0) { sid2_die("EOF code: invalid module size\n"); return; }

 char *module=malloc(sid2fs); if(!sid2fs) { sid2_die("EOF code: malloc failed\n"); return; }
 FILE *f=fopen(input_filename, "r"); if(!f) { sid2_die("EOF code: fopen failed\n"); return; }
 fread(module, sid2fs, 1, f); fclose(f);

 sid2_st=sidplay2_st_create((const uint_least8_t *)module, sid2fs);
 if(sid2_st) { dlog(9,"sidplay2_st_create!=false"); dlog(9,(char *)sidplay2_error(sid2c)); }

 ret=sidplay2_st_selectsong(sid2_st, 0);
 sprintf(lmsg, "sidplay2_st_selectsong: %d", ret); dlog(9,lmsg);

 ret=sidplay2_load(sid2c, sid2_st);
 sprintf(lmsg, "sidplay2_load: %d", ret); dlog(9,lmsg);


 sidplay2_config_get(sid2c, &sid2conf);

 sid2conf.frequency=44100;
// sid2conf.frequency=22050;
 sid2conf.sampleFormat=SID2_LITTLE_SIGNED;
 sid2conf.optimisation=1;
 sid2conf.precision=16;
if(1) sid2conf.playback=sid2_mono; else sid2conf.playback=sid2_stereo;
 sid2conf.forceDualSids=false;
 sid2conf.sidModel=SID2_MODEL_CORRECT;
 void *rs=sidplay2_newresid(1, sid2conf.frequency);
 sid2conf.sidEmulation = rs; if(!rs) dlog(9, "setting sidEmulation failed");

 ret=sidplay2_config_set(sid2c, &sid2conf);

//if(ret<0) { sid2_die((char *)sidplay2_error(sid2c)); return; }
if(ret<0) { dlog(9,(char *)sidplay2_error(sid2c)); }

 free(module);
// if(ret!=0 || ret2!=0)
// if(ret==0 || ret2==0)
 {
//sprintf(lmsg, "sid2 fs:%d, ret0:%d, ret:%d, ret2:%d", sid2fs, ret0, ret, ret2); dlog(9,lmsg);
// sid2_die("EOF code: sid2_load failed\n");
// return;
 }

 sfmt=AV_SAMPLE_FMT_S16;
      if(sfmt==AV_SAMPLE_FMT_U8 || sfmt==AV_SAMPLE_FMT_U8P){ sformat.bits=8; }
 else if(sfmt==AV_SAMPLE_FMT_S16 || sfmt==AV_SAMPLE_FMT_S16P){ sformat.bits=16; }
 else if(sfmt==AV_SAMPLE_FMT_S32 || sfmt==AV_SAMPLE_FMT_S32P){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_FLT || sfmt==AV_SAMPLE_FMT_FLTP){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_DBL || sfmt==AV_SAMPLE_FMT_DBLP){ sformat.bits=64; }
 else { sid2_die("EOF code: Unsupported format\n"); return; }

 sid2cur=-1;
// sid2dur=sid2Info_GetDuration(sid2I, sid2Info_GetDefaultSong(sid2I))/1000.0;

 sformat.channels=(sid2conf.playback==sid2_stereo?2:1);
 sformat.rate=sid2conf.frequency;
#ifdef LAO
 sformat.byte_format=AO_FMT_NATIVE; sformat.matrix=0;
#endif

 adev_setp(sformat.bits, &sformat.channels, &sformat.rate);
sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels, 0); cs(lmsg);

 sid2_update_info(input_filename);

 sid2state=3; sid2i=1;
 sid2_play();
}
void oscp_sid2_check_input();
void oscp_sid2_parse(char *a);
/////################# check messages
void sid2_check_input() { lf("sid2_check_input"); static char lmsg[4096]; static int rc;
 static fd_set sid2fsa;  static struct timeval sid2tv,*tvp;
 if(sid2state==4) { tvp=NULL; }
 else { sid2tv.tv_sec=0; sid2tv.tv_usec=1; tvp=&sid2tv; }

 FD_ZERO(&sid2fsa); if(parent_said>=0) FD_SET(parent_said, &sid2fsa);

 rc=select(FD_SETSIZE, &sid2fsa, NULL, NULL, tvp); if(rc==0) return;
 else if(rc<0) { sprintf(lmsg, "Info: sid2ci rc<0, %d, %s\n", errno, strerror(errno)); cs(lmsg); }
 else if(rc>0) // check if input is available
 { fdreadline(parent_said, lmsg, 1020); oscp_sid2_parse(lmsg); }
}
/////###################
void sid2_pause_toggle() { lf("sid2_pause_toggle");
      if(sid2state==3) { sid2state=4; cs("===== PAUSE =====\n");
  do { oscp_sid2_check_input(); } while(sid2state==4);
 }
 else if(sid2state==4) { sid2state=3; cs("===== RESUME =====\n"); }
 else cs("Info: pause called but sid2state != (3 || 4)"); }

/////##############
void sid2_checksd() {}
/////##############
void sid2_checkmd() {}

/////###########
void sid2_play() { lf("sid2_play"); dlog(9, "sid2_play");
 if(sid2state!=3) { cs("EOF code: can't play, not inited\n"); return; }
 float ot1=-7.0, ts=((float)sid2_dbuf_size)/sformat.rate/sformat.channels/2.0;

 static char lmsg[128]; void *buf[1];

 while(sid2state==3 && sidplay2_play(sid2c, sid2_dbuf, sid2_dbuf_size)>0 && sid2cur<=sid2dur)
 {
  sid2cur+=ts; if((int)sid2cur!=(int)ot1)
  { ot1=sid2cur; sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0\n", sid2cur); cs(lmsg); }

  buf[0]=sid2_dbuf; adev_out(buf, sid2_dbuf_size/2/sformat.channels, sformat.channels, sfmt);

  sid2i++; sid2_check_input();
 }

 adev_close();
 sid2_die("EOF code: finished decoding\n");
 lf("sid2_play:finished");
}

/////########################
void oscp_sid2_load(char *fn) { lf("oscp_sid2_load");
 sid2_load(fn); //sid2_play();
 lf("oscp_sid2_load:finished");
}
/////#####################
void oscp_sid2_seek(int a) { lf("oscp_sid2_seek"); if(a<=0) return;
 float ts=((float)sid2_dbuf_size)/sformat.rate/sformat.channels/2.0; float s0=sid2cur; int db=0;
 while((db=sidplay2_play(sid2c, sid2_dbuf, sid2_dbuf_size))>0 && sid2cur<s0+a) { sid2cur+=ts; }
}
/////########################
void oscp_sid2_parse(char *a) { lf("oscp_sid2_parse"); char lmsg[4096];
// sprintf(lmsg,"ofparse start"); dlog(9,lmsg);
 int l=strlen(a); if(l<=0) { sprintf(lmsg,"ofparse: l==%d ?", l); dlog(9,lmsg); return; }
 if(strcmp(a,"q")==0 || strcmp(a,"quit")==0) { sid2_end(); exit(0);
 } else if(strncmp(a,"loadfile ",9)==0) { oscp_sid2_load(a+9);
 } else if(strncmp(a,"load ",5)==0) { oscp_sid2_load(a+5);
 } else if(strncmp(a,"pause",5)==0) { sid2_pause_toggle();
 } else if(strncmp(a,"get_property length",19)==0) {
  sprintf(lmsg, "ANS_length=%.1f\n", sid2dur); cs(lmsg);
 } else if(strncmp(a,"get_property time_pos",19)==0 || strncmp(a,"get_time_pos",12)==0) {
//  sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0 0 0\n", sid2cur); cs(lmsg);
 } else if(strncmp(a,"seek ",5)==0) { int o1=0;
  if(a[5]==*"-") { sscanf(a+6, "%d", &o1); o1*=-1; }
  else if(a[5]==*"+") { sscanf(a+6, "%d", &o1); }
  else { sscanf(a+5, "%d", &o1); }
  oscp_sid2_seek(o1);
 } else if(strncmp(a,"vis ",4)==0) { sscanf(a+4, "%d", &vis);
 } else if(strncmp(a,"lfdump",5)==0) { lfdump();
 } else { sprintf(lmsg, "ofparse unhandled: %s", a); dlog(9, lmsg); }
 lf("oscp_sid2_parse:finished");
}
/////####################### check stdin
void oscp_sid2_check_input() { lf("oscp_sid2_check_input");
 char lmsg[4096]; int i, rc; fd_set fsa;
 struct timeval *tvp=NULL; //, tv; tv.tv_sec=0; tv.tv_usec=10000*99; // i=(int)time(NULL);
 usleep(10000); FD_ZERO(&fsa); if(parent_said>=0) FD_SET(parent_said, &fsa);
 rc=select(FD_SETSIZE, &fsa, NULL, NULL, tvp); // block till input

 if(rc<0) { if(0) { exit(EXIT_FAILURE); } else usleep(10000*99); }
 else if(rc==0) { }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==parent_said) { fdreadline(i, lmsg, 1020); oscp_sid2_parse(lmsg); }
 }
 lf("oscp_sid2_check_input:finished");
}
////################
int oscp_sid2_main() { lf("oscp_sid2_main");
 cs("oscp-slave sidplay2 2.1.1\n");
 lf("oscp_sid2_main:sid2_init");
 sid2_init();
 lf("oscp_sid2_main:while");
 while(1) { oscp_sid2_check_input(); }
 lf("oscp_sid2_main:sid2_end");
 sid2_end();
 lf("oscp_sid2_main:finished");
 cs("EOF code: exit\n");
// lfdump();
 return 0;
}
